# Copyright 2020 Andrzej Cichocki

# This file is part of Leytonium.
#
# Leytonium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Leytonium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Leytonium.  If not, see <http://www.gnu.org/licenses/>.

'Instant mortgage statement.'
from argparse import ArgumentParser
from aridity.config import ConfigCtrl
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, timedelta
from functools import cache

maxmonth = timedelta(31)
oneday = timedelta(1)

def _days(start, end):
    return int((end - start) / oneday)

@cache
def _yearlen(year):
    return _days(date(year, 1, 1), date(year + 1, 1, 1))

@dataclass
class Period:

    firstday: object
    payment: object
    rate: object

def main():
    config = ConfigCtrl().loadappconfig(main, 'mortcalc.arid')
    parser = ArgumentParser()
    parser.add_argument('--short', action = 'store_true')
    parser.parse_args(namespace = config.cli)
    today = date.today()
    short = config.short
    ring = []
    balance = config.anchor
    periods = [Period(date.fromisoformat(k), c.payment, c.rate / 100) for k, c in -config.firstday]
    mark = periods[0].firstday
    for p, q in zip(periods, periods[1:]):
        assert p.firstday < q.firstday
    value = config.value
    holidays = set(map(date.fromisoformat, config.holiday))
    overpays = {date.fromisoformat(k): v for k, v in -config.overpay}
    for period, nextperiod in zip(periods, [*periods[1:], None]):
        while nextperiod is None or mark < nextperiod.firstday:
            nextmark = (mark + maxmonth).replace(day = 1)
            if short and mark <= today < nextmark:
                for t in ring:
                    print(*t)
                return
            dayrate = period.rate / _yearlen(mark.year)
            paymark = mark
            while paymark.weekday() > 4 or paymark in holidays:
                paymark += oneday
            payments = defaultdict(int)
            payments[paymark] = period.payment
            for m, p in overpays.items():
                if m.year == mark.year and m.month == mark.month:
                    payments[m] += p
            payments[nextmark] = 0
            interest = 0
            cursor = mark
            for m, p in sorted(payments.items()):
                for d in range(_days(cursor, m)):
                    if balance <= 0:
                        return
                    interest += balance * dayrate
                    b = balance + interest
                    if not short:
                        print(cursor + timedelta(d), interest, b, b / value * 100, balance)
                cursor = m
                balance -= p
            balance += round(interest, 2)
            ring.append((nextmark - oneday, b))
            del ring[:-2]
            mark = nextmark

if '__main__' == __name__:
    main()



# Copyright 2020 Andrzej Cichocki

# This file is part of Leytonium.
#
# Leytonium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Leytonium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Leytonium.  If not, see <http://www.gnu.org/licenses/>.

'Shell for cron.'
from datetime import datetime
from lagoon.text import bash, lsof
from pathlib import Path
from subprocess import DEVNULL
import shlex, sys

def main():
    assert '-c' == sys.argv[1]
    script = sys.argv[2]
    name = Path(shlex.split(script)[0]).name
    logpath = Path('var', 'log', f"{name}.log")
    inuse = 'aw' in lsof._F.a[::DEVNULL](logpath, check = False).stdout.splitlines()
    with logpath.open('a') as f:
        def log(text):
            print(datetime.now(), text, file = f)
        if inuse:
            log('Previous instance still running.')
        else:
            bash._lc[:f:f](sys.stdin.read() + script) # XXX: Is login shell needed in .xsessionrc case?
            log('Normal end.')

if '__main__' == __name__:
    main()



# Copyright 2020 Andrzej Cichocki

# This file is part of Leytonium.
#
# Leytonium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Leytonium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Leytonium.  If not, see <http://www.gnu.org/licenses/>.

'Build free-threaded Python Docker image.'
from aridity.config import ConfigCtrl
from foyndation import solo
from lagoon.text import docker
from lagoon.url import URL
from pathlib import Path
from tempfile import TemporaryDirectory
import re

def main():
    config = ConfigCtrl().loadappconfig(main, 'nogil.arid')
    dockerfile = URL.text(config.url)()
    j = solo(re.finditer('--with-ensurepip', dockerfile)).end()
    dockerfile = f"{dockerfile[:j]} --disable-gil{dockerfile[j:]}"
    with TemporaryDirectory() as context:
        Path(context, 'Dockerfile').write_text(dockerfile)
        docker.build[print]('-t', config.tag, context)

if '__main__' == __name__:
    main()



# Copyright 2020 Andrzej Cichocki

# This file is part of Leytonium.
#
# Leytonium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Leytonium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Leytonium.  If not, see <http://www.gnu.org/licenses/>.

'Play a sound when a long-running child of shell terminates.'
from argparse import ArgumentParser
from aridity.config import ConfigCtrl
from foyndation import inf, innerclass
from lagoon.program import partial
from lagoon.text import pgrep
from pathlib import Path
from subprocess import CalledProcessError
import os, time

class TaskDing:

    def __init__(self, config):
        self.always_interactive = set(config.always.interactive)
        self.pgrep = pgrep[partial]('-P', config.shpidstr)
        self.sleep_time = float(config.sleep.time)
        self.sound_path = Path(config.sound.path)
        self.threshold = config.threshold

    @innerclass
    class Child:

        armed = False

        def __init__(self, mark):
            self.mark = mark

        def tick(self, now, pid):
            if self.mark <= now:
                self.mark = inf
                try:
                    self.armed = Path(f"/proc/{pid}/comm").read_text().rstrip() not in self.always_interactive
                except (FileNotFoundError, ProcessLookupError):
                    pass

        def fire(self):
            from lagoon.text import paplay
            if self.armed and self.sound_path.exists():
                if (pid := os.fork()):
                    return pid
                paplay[exec](self.sound_path)

    def run(self):
        children = {}
        soundpids = set()
        while True:
            nowchildren = {}
            now = time.time()
            try:
                with self.pgrep as stdout:
                    for line in stdout:
                        nowchildren[int(line)] = self.Child(now + self.threshold)
            except CalledProcessError:
                break
            for pid in soundpids - nowchildren.keys():
                os.waitpid(pid, 0)
            soundpids &= nowchildren.keys()
            for pid in children.keys() - nowchildren.keys():
                q = children.pop(pid).fire()
                if q is not None:
                    soundpids.add(q)
            for pid, child in nowchildren.items():
                if pid not in children:
                    children[pid] = child
            for pid, child in children.items():
                child.tick(now, pid)
            time.sleep(self.sleep_time) # FIXME LATER: I suspect keyboard interrupt can kill script when not asleep.

def main():
    if 'SSH_CLIENT' in os.environ:
        return
    config = ConfigCtrl().loadappconfig(main, 'taskding.arid')
    parser = ArgumentParser()
    parser.add_argument('shpidstr')
    parser.parse_args(namespace = config.cli)
    TaskDing(config).run()

if '__main__' == __name__:
    main()
